/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.solr.analytics.util.function;

import java.util.Objects;
import java.util.function.Consumer;

/**
 * Represents an operation that accepts a single {@code boolean}-valued argument and returns no
 * result. This is the primitive type specialization of {@link Consumer} for {@code boolean}. Unlike
 * most other functional interfaces, {@code IntConsumer} is expected to operate via side-effects.
 *
 * <p>This is a <a href="package-summary.html">functional interface</a> whose functional method is
 * {@link #accept(boolean)}.
 *
 * @see Consumer
 */
@FunctionalInterface
public interface BooleanConsumer {

  /**
   * Performs this operation on the given argument.
   *
   * @param value the input argument
   */
  void accept(boolean value);

  /**
   * Returns a composed {@code DoubleConsumer} that performs, in sequence, this operation followed
   * by the {@code after} operation. If performing either operation throws an exception, it is
   * relayed to the caller of the composed operation. If performing this operation throws an
   * exception, the {@code after} operation will not be performed.
   *
   * @param after the operation to perform after this operation
   * @return a composed {@code DoubleConsumer} that performs in sequence this operation followed by
   *     the {@code after} operation
   * @throws NullPointerException if {@code after} is null
   */
  default BooleanConsumer andThen(BooleanConsumer after) {
    Objects.requireNonNull(after);
    return (boolean t) -> {
      accept(t);
      after.accept(t);
    };
  }
}
